#define __NO_VERSION__
#include <linux/module.h>

#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/locks.h>
#include <asm/uaccess.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static int errno;

#include "arcfs.h"
#include "mblock.h"


#define ARCBSIZE 512
#define ARCBSBITS 9



//static int arcfs_major = 0;
//static struct file_operations arcfs_fops;
static struct super_operations arcfs_sops;
static struct file_operations  arcfs_file_operations;
static struct inode_operations arcfs_file_inode_operations;

static int inline min(int a, int b)
{
	return a<b ? a : b;
}

struct super_block *arcfs_read_super(struct super_block *s, void *data,
				     int silent)
{
  kdev_t dev = s->s_dev;
  char *arcfile;
  MBlock mem;
  struct arcfs_module *mod;
  struct arcfs_super_block *asb;

  if(!data || !*(char *)data)
    return NULL;

  PRINTK(("option=[%s]\n", (char *)data));

  if((arcfile = strstr((char *)data, "arc=")) == NULL)
    return NULL;
  arcfile += 4;
  if((mod = arcfs_find_module(arcfile)) == NULL)
    {
      PRINTK(("no arc module\n"));
      return NULL;
    }

  PRINTK(("arcfs_read_super(%s) dev=%d\n", arcfile, s->s_dev));
  lock_super(s);

  set_blocksize(dev, ARCBSIZE);
  s->s_blocksize = ARCBSIZE;
  s->s_blocksize_bits = ARCBSBITS;
  s->s_flags |= MS_RDONLY;
  s->s_op = &arcfs_sops;

  mblock_init(&mem);
  s->u.generic_sbp = asb = (struct arcfs_super_block *)
    mblock_alloc(&mem, sizeof(struct arcfs_super_block));
  if(asb == NULL)
    goto error1;

  asb->mem = mem;	/* Save the pointer of self memory pool */

  asb->module = mod;

  asb->root = arcfs_alloc_entry(s, arcfile, strlen(arcfile));
  asb->root->sub = NULL;
  asb->root->mode = 0555 | S_IFDIR;
  asb->root->size = 512;
  asb->arcfile = asb->root->name;

  /* Install archive file */
  {
    int ret;
    mm_segment_t fs;
    fs = get_fs();
    set_fs(get_ds());
    ret = mod->install_arcfs(s);
    set_fs(fs);
    if(ret < 0)
      goto error1;
  }

  s->s_root = d_alloc_root(iget(s, ARCENT2INO(asb->root)), NULL);
  if(!s->s_root)
    goto error1;

  MOD_INC_USE_COUNT;
  unlock_super(s);
  return s;

 error1:
  unlock_super(s);
  mblock_free((MBlock *)&asb->mem);
  // error0:
  return NULL; 
}

extern struct file_system_type arc_fs_type;
extern int init_arc_fs(void);
int init_module(void)
{
  PRINTK(("init module()\n"));
  return init_arc_fs();
}

void cleanup_module(void)
{
  PRINTK(("cleanup_module()\n"));
  unregister_filesystem(&arc_fs_type);
}

static void
arcfs_read_inode(struct inode *i)
{
  struct arcfs_entry *entry;

  entry = INO2ARCENT(i->i_ino);
  i->i_op = &arcfs_file_inode_operations;
  i->i_mode = entry->mode;
  i->i_size = entry->size;
  i->i_uid = entry->uid;
  i->i_gid = entry->gid;
  i->i_mtime = entry->mtime;
  i->i_atime = entry->atime;
  i->i_ctime = entry->ctime;
  i->i_nlink = 1;		/* Hard to decide.. */

  if(S_ISDIR(i->i_mode) && i->i_size == 0)
    i->i_size = ARCBSIZE;
}

static void
arcfs_put_super(struct super_block *s)
{
  struct arcfs_super_block *asb;

  PRINTK(("arcfs_put_super()\n"));

  asb = (struct arcfs_super_block *)s->u.generic_sbp;
  asb->module->uninstall_arcfs(s);
  MOD_DEC_USE_COUNT;
  mblock_free((MBlock *)&asb->mem);
}

struct dentry *
arcfs_lookup(struct inode *dir, struct dentry *dentry)
{
  const char *name;		/* got from dentry */
  int len;
  struct arcfs_entry *entry;
  int res;

  PRINTK(("arcfs_lookup(%s,%.*s)\n", dentry->d_name.name, (int)dentry->d_name.len, dentry->d_iname));

  res = -EBADF;
  if (!dir || !S_ISDIR(dir->i_mode))
    goto out;

  res = 0;
  name = dentry->d_name.name;
  len = dentry->d_name.len;
  entry = INO2ARCENT(dir->i_ino);
  entry = entry->sub;
  while(entry)
    {
      if(strncmp(name, entry->name, len) == 0 && entry->name[len] == '\0')
	break;
      entry = entry->next;
  }
  if(entry == NULL)
    return ERR_PTR(-ENOENT);
  if(entry)
    {
      struct inode *inode;
      if((inode = iget(dir->i_sb, ARCENT2INO(entry))) == NULL) {
	res = -EACCES;
      } else {
	d_add(dentry, inode);
      }
    }

 out:
  return ERR_PTR(res);
}

static int
arcfs_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
  struct inode *i = filp->f_dentry->d_inode;
  int stored = 0;
  unsigned long ino;
  struct arcfs_entry *entry, *base;
  int n;

  PRINTK(("arcfs_readdir(), ino=%lx, pos=%lu\n", i->i_ino, (unsigned long)filp->f_pos));
  i->i_atime = CURRENT_TIME;
  base = INO2ARCENT(i->i_ino);

  if(filp->f_pos == 0)
    {
      if (filldir(dirent, ".", 1, filp->f_pos, ARCENT2INO(base)) < 0)
	return stored;
      filp->f_pos = 256;
    }
  if(filp->f_pos == 256)
    {
      if (filldir(dirent, "..", 2, filp->f_pos, ARCENT2INO(base)) < 0)
	return stored;
      filp->f_pos = 512;
    }

  n = filp->f_pos / 256 - 2;
  entry = base->sub;
  while(n-- > 0 && entry)
    entry = entry->next;
  while(entry)
    {
      ino = ARCENT2INO(entry);
      if (filldir(dirent, entry->name, strlen(entry->name), filp->f_pos, ino) < 0)
	return stored;
      filp->f_pos += 256;
      entry = entry->next;
    }

  return stored;
}

static int arcfs_open(struct inode *inode, struct file *filp)
{
  struct arcfs_entry *entry;
  struct arcfs_super_block *asb;
  struct arc_file *arc_filp;
  mm_segment_t fs;

  PRINTK(("arcfs_open()\n"));
  inode->i_atime = CURRENT_TIME;
  entry = INO2ARCENT(inode->i_ino);
  filp->private_data = NULL;

  if(S_ISDIR(entry->mode))
      return 0;

  asb = ENTRY_ARCFS(entry);

  fs = get_fs();
  set_fs(get_ds());
  arc_filp = asb->module->arc_open(entry);
  set_fs(fs);

  if(IS_ERR(arc_filp))
    return PTR_ERR(arc_filp);

  arc_filp->entry = entry;
  filp->private_data = arc_filp;
  return 0;
}

#if 0
static ssize_t arcfs_read(struct file *filp, char *buf, size_t len,
			  loff_t *off)
{
  struct arc_file *arc_filp;
  struct arcfs_module *mod;
  ssize_t n;

  arc_filp = (struct arc_file *)filp->private_data;
  if(!arc_filp)
    return 0;
  mod = ENTRY_ARCFS(arc_filp->entry)->module;

  if(*off != arc_filp->pos)
    {
      loff_t p;
      if((p = mod->arc_seek(arc_filp, *off)) < 0)
	return p;
      *off = arc_filp->pos;
    }
  if(*off >= arc_filp->size)
    return 0;
  if(arc_filp->pos + len > arc_filp->size)
    len = arc_filp->size - arc_filp->pos;
  n = mod->arc_read(arc_filp, buf, len);
  if(n < 0)
    return n;
  *off += n;
  return n;
}
#endif

static int
arcfs_readpage(struct file * filp, struct page * page)
{
  struct dentry *dentry = filp->f_dentry;
  struct inode *inode = dentry->d_inode;
  unsigned long buf;
  unsigned long offset, avail, readlen;
  int result = -EIO;

  PRINTK(("arcfs_readpage()\n"));
  inode->i_atime = CURRENT_TIME;

  atomic_inc(&page->count);
  set_bit(PG_locked, &page->flags);

  buf = page_address(page);
  clear_bit(PG_uptodate, &page->flags);
  clear_bit(PG_error, &page->flags);
  offset = page->offset;

  if(offset < inode->i_size)
    {
      struct arc_file *arc_filp;
      struct arcfs_module *mod;
      ssize_t n;
      mm_segment_t fs;

      avail = inode->i_size-offset;
      readlen = min(avail, PAGE_SIZE);

      arc_filp = (struct arc_file *)filp->private_data;
      mod = ENTRY_ARCFS(arc_filp->entry)->module;
      fs = get_fs();
      set_fs(get_ds());
      if(offset != arc_filp->pos)
	{
	  n = mod->arc_seek(arc_filp, offset);
	  if(n < 0)
	    {
	      set_fs(fs);
	      result = n;
	      goto ret;
	    }
	}

      n = mod->arc_read(arc_filp, (char *)buf, readlen);
      set_fs(fs);

      if(n == readlen)
	{
	  if (readlen < PAGE_SIZE)
	      memset((void *)(buf+readlen),0,PAGE_SIZE-readlen);
	  set_bit(PG_uptodate, &page->flags);
	  result = 0;
	}
      else if(n < 0)
	result = n;
    }

 ret:
  if(result)
    {
      set_bit(PG_error, &page->flags);
      memset((void *)buf, 0, PAGE_SIZE);
    }

  clear_bit(PG_locked, &page->flags);
  wake_up(&page->wait);
  free_page(buf);

  return result;
}

static int arcfs_release(struct inode *inode, struct file *filp)
{
  PRINTK(("arcfs_release()\n"));

  if(filp->private_data)
    {
      struct arc_file *arc_filp;
      struct arcfs_module *mod;
      arc_filp = (struct arc_file *)filp->private_data;
      mod = ENTRY_ARCFS(arc_filp->entry)->module;
      mod->arc_close(arc_filp);
    }
  return 0;
}

static int arcfs_readlink(struct dentry *dentry, char *buf, int bufsiz)
{
  struct arcfs_entry *entry;
  int len;

  entry = INO2ARCENT(dentry->d_inode->i_ino);
  if(!S_ISLNK(entry->mode))
    return -EINVAL;
  if(bufsiz < entry->size)
    len = bufsiz;
  else
    len = entry->size;
  copy_to_user(buf, entry->symlink, len);
  return len;
}

static struct super_operations arcfs_sops =
{
  arcfs_read_inode,	/* read inode */
  NULL,			/* write inode */
  NULL,			/* put inode */
  NULL,			/* delete inode */
  NULL,			/* notify change */
  arcfs_put_super,	/* put super */
  NULL,			/* write super */
  NULL,			/* statfs */
  NULL			/* remount */
};

static struct file_operations  arcfs_file_operations =
{
	NULL,			/* lseek - default */
	generic_file_read,
	NULL,			/* write - bad */
	arcfs_readdir,		/* readdir */
	NULL,			/* poll - default */
	NULL,			/* ioctl */
	generic_file_mmap,	/* mmap */
	arcfs_open,		/* open */
	NULL,			/* flush */
	arcfs_release,		/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

static struct inode_operations arcfs_file_inode_operations = {
	&arcfs_file_operations, /* default_file_ops */
	NULL,			/* create */
	arcfs_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	arcfs_readlink,		/* readlink */
	NULL,			/* follow_link */
	arcfs_readpage,		/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap -- not really */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL,			/* smap */
};

//MODULE_PARM(arcfs_major, "i");
